#!/usr/bin/env python

# Copyright (C) 2002-2003 Gideon May (gideon@computer.org)
#
# Permission to copy, use, sell and distribute this software is granted
# provided this copyright notice appears in all copies.
# Permission to modify the code and to distribute modified code is granted
# provided this copyright notice appears in all copies, and a notice
# that the code was modified is included with the copyright notice.
#
# This software is provided "as is" without express or implied warranty,
# and with no claim as to its suitability for any purpose.

# -*-c++-*- OpenSceneGraph - Copyright (C) 1998-2003 Robert Osfield 
#
# This application is open source and may be redistributed and/or modified   
# freely and without restriction, both in commericial and non commericial applications,
# as long as this copyright notice is maintained.
# 
# This application is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#

import sys
from PyOSG import *
from PyOSG import osg
from PyOSG import osgDB
from PyOSG import osgUtil
from PyOSG import osgGA
from PyOSG import osgProducer
import OpenGL

# Creates a stateset which contains a 1D texture which is populated by contour banded color
# this is then used in conjunction with TexGen to create contoured models, either in 
# object linear coords - like contours on a map, or eye linear which contour the distance from
# the eye. An app callback toggles between the two tex gen modes.
def create1DTextureStateToDecorate(loadedModel):
    bs = loadedModel.getBound()

    image = osg.Image()

    noPixels = 1024

    # allocate the image data, noPixels x 1 x 1 with 4 rgba floats - equivilant to a Vec4!
    image.allocateImage(noPixels,1,1,OpenGL.GL.GL_RGBA,OpenGL.GL.GL_FLOAT)
    image.setInternalTextureFormat(OpenGL.GL.GL_RGBA)


    colorbands = []
    colorbands.append(osg.Vec4(0.0,0.0,0.0,1.0))
    colorbands.append(osg.Vec4(1.0,0.0,0.0,1.0))
    colorbands.append(osg.Vec4(1.0,1.0,0.0,1.0))
    colorbands.append(osg.Vec4(0.0,1.0,0.0,1.0))
    colorbands.append(osg.Vec4(0.0,1.0,1.0,1.0))
    colorbands.append(osg.Vec4(0.0,0.0,1.0,1.0))
    colorbands.append(osg.Vec4(1.0,0.0,1.0,1.0))
    colorbands.append(osg.Vec4(1.0,1.0,1.0,1.0))

    nobands = len(colorbands)
    delta = float(nobands)/noPixels
    pos = 0.0

    # fill in the image data
    for v in image.dataAsVec4:
        # p = int(pos)
        # r = pos - p
        # color = colorbands[p] * (1.0 - r)
        # if p+1 < len(colorbands) :
        #     color += colorbands[p + 1] * r
        color = colorbands[int(pos)]
        v.set(color[0], color[1], color[2], color[3])
        pos += delta

    texture = osg.Texture1D()
    texture.setWrap(osg.Texture1D.WRAP_S,osg.Texture1D.MIRROR)
    texture.setFilter(osg.Texture1D.MIN_FILTER,osg.Texture1D.LINEAR)
    texture.setImage(image)

    zBase = bs.center()._z-bs.radius()
    zScale = 2.0/bs.radius()

    texgen = osg.TexGen()
    texgen.setDataVariance(osg.Object.DYNAMIC)
    texgen.setMode(osg.TexGen.OBJECT_LINEAR)
    texgen.setMode(osg.TexGen.EYE_LINEAR)
    texgen.setPlane(osg.TexGen.S, osg.Plane(osg.Vec4(0.0,0.0,zScale,-zBase)))

    material = osg.Material()

    stateset = osg.StateSet()

    stateset.setTextureAttribute(0,texture,osg.StateAttribute.OVERRIDE)
    stateset.setTextureMode(0,OpenGL.GL.GL_TEXTURE_1D,osg.StateAttribute.ON|osg.StateAttribute.OVERRIDE)
    stateset.setTextureMode(0,OpenGL.GL.GL_TEXTURE_2D,osg.StateAttribute.OFF|osg.StateAttribute.OVERRIDE)
    #stateset.setTextureMode(0,OpenGL.GL.GL_TEXTURE_3D,osg.StateAttribute.OFF|osg.StateAttribute.OVERRIDE)

    stateset.setTextureAttribute(0,texgen,osg.StateAttribute.OVERRIDE)
    stateset.setTextureMode(0,OpenGL.GL.GL_TEXTURE_GEN_S,osg.StateAttribute.ON|osg.StateAttribute.OVERRIDE)

    stateset.setAttribute(material,osg.StateAttribute.OVERRIDE)

    return stateset

# An app callback which alternates the tex gen mode between object linear and eye linear to illustrate what differences it makes.
class AnimateStateCallback(osg.NodeCallback):
    previousTime = 0.0
    state = 0
    def __init__(self):
        self._self = self
        osg.NodeCallback.__init__(self)

    def animateState(self, stateset,time):
        # here we simply get any existing texgen, and then increment its
        # plane, pushing the R coordinate through the texture.
        attribute = stateset.getTextureAttribute(0,osg.StateAttribute.TEXGEN)
        texgen = attribute.asTexGen()
        #-- print texgen
        if texgen:
            timeInterval = 1.0

            #self.previousTime = time
            while time>AnimateStateCallback.previousTime+timeInterval:
                AnimateStateCallback.previousTime += timeInterval
                AnimateStateCallback.state = not AnimateStateCallback.state

            #-- print AnimateStateCallback.previousTime, AnimateStateCallback.state

            if AnimateStateCallback.state == 1:
                texgen.setMode(osg.TexGen.OBJECT_LINEAR)
            else:
                texgen.setMode(osg.TexGen.EYE_LINEAR)
        

    def apply(self, node, nv):
        stateset = node.getStateSet()
        if stateset and nv.getFrameStamp():
            # we have an exisitng stateset, so lets animate it.
            self.animateState(stateset,nv.getFrameStamp().getReferenceTime())

        # note, callback is repsonsible for scenegraph traversal so
        # should always include call the traverse(node,nv) to ensure 
        # that the rest of cullbacks and the scene graph are traversed.
        self.traverse(node,nv)

def main(argv):
    # use an ArgumentParser object to manage the program arguments.
    arguments =  osg.ArgumentParser(argv)

    # set up the usage document, in case we need to print out how to use this program.
    arguments.getApplicationUsage().setDescription(arguments.getApplicationName()+" is the example which demonstrates use of 1D textures.");
    arguments.getApplicationUsage().setCommandLineUsage(arguments.getApplicationName()+" [options] filename ...")
    arguments.getApplicationUsage().addCommandLineOption("-h or --help","Display this information")

    # construct the viewer.
    viewer = osgProducer.Viewer(arguments)

    # set up the value with sensible default event handlers.
    viewer.setUpViewer(osgProducer.Viewer.STANDARD_SETTINGS)

    # get details on keyboard and mouse bindings used by the viewer.
    viewer.getUsage(arguments.getApplicationUsage())

    # if user request help write it out to cout.
    if arguments.read("-h") or arguments.read("--help"):
        arguments.getApplicationUsage().write()
        return 1

    # any option left unread are converted into errors to write out later.
    arguments.reportRemainingOptionsAsUnrecognized()

    # report any errors if they have occured when parsing the program aguments.
    if arguments.errors():
        arguments.writeErrorMessages()
        return 1
    
    if arguments.argc() <= 1:
        arguments.getApplicationUsage().write(osg.ApplicationUsage.COMMAND_LINE_OPTION)
        return 1

    # load the images specified on command line
    loadedModel = osgDB.readNodeFiles(arguments)

    if loadedModel:

        stateset = create1DTextureStateToDecorate(loadedModel)
        if not stateset:
            print "Error: failed to create 1D texture state."
            return 1


        loadedModel.setStateSet(stateset)
        loadedModel.setUpdateCallback(AnimateStateCallback())

        # add model to viewer.
        viewer.setSceneData( loadedModel )

        # create the windows and run the threads.
        viewer.realize()

        while not viewer.done():
            # wait for all cull and draw threads to complete.
            viewer.sync()

            # update the scene by traversing it with the the update visitor which will
            # call all node update callbacks and animations.
            viewer.update()
             
            # fire off the cull and draw traversals of the scene.
            viewer.frame()
        
        # wait for all cull and draw threads to complete before exit.
        viewer.sync()

    else:
        print arguments.getApplicationUsage().getCommandLineUsage()
        return 0

    return 0

if __name__ == "__main__":
    main(sys.argv)
