#!/usr/bin/env python

# Copyright (C) 2002-2003 Gideon May (gideon@computer.org)
#
# Permission to copy, use, sell and distribute this software is granted
# provided this copyright notice appears in all copies.
# Permission to modify the code and to distribute modified code is granted
# provided this copyright notice appears in all copies, and a notice
# that the code was modified is included with the copyright notice.
#
# This software is provided "as is" without express or implied warranty,
# and with no claim as to its suitability for any purpose.

# -*-c++-*- OpenSceneGraph - Copyright (C) 1998-2003 Robert Osfield 
#
# This application is open source and may be redistributed and/or modified   
# freely and without restriction, both in commericial and non commericial applications,
# as long as this copyright notice is maintained.
# 
# This application is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#

import sys
from PyOSG import Producer
from PyOSG import osg
from PyOSG import osgDB
from PyOSG import osgUtil
from PyOSG import osgGA
from PyOSG import osgText
from PyOSG import osgProducer
from OpenGL.GL import *

#####################################################################/
#
# filter wall and animation callback.
#

class FilterCallback( osg.NodeCallback):
    def __init__(self, texture, text, delay=1.0):
        self._texture = texture
        self._text = text
        self._delay = delay
        self._currPos = 0
        self._prevTime = 0.0
    
        self._minFilterList = []
        self._magFilterList = []
        self._textList = []
    
    
        # start with a mip mapped mode to ensure that
        self._minFilterList.append(osg.Texture2D.LINEAR_MIPMAP_LINEAR)
        self._magFilterList.append(osg.Texture2D.LINEAR)
        self._textList.append("Tri-linear mip mapping (default filtering)\nsetFilter(MIN_FILTER,LINEAR_MIP_LINEAR)\nsetFilter(MAG_FILTER,LINEAR)")

        self._minFilterList.append(osg.Texture2D.NEAREST)
        self._magFilterList.append(osg.Texture2D.NEAREST)
        self._textList.append("Nearest filtering\nsetFilter(MIN_FILTER,NEAREST)\nsetFilter(MAG_FILTER,NEAREST)")
        
        self._minFilterList.append(osg.Texture2D.LINEAR)
        self._magFilterList.append(osg.Texture2D.LINEAR)
        self._textList.append("Linear filtering\nsetFilter(MIN_FILTER,LINEAR)\nsetFilter(MAG_FILTER,LINEAR)")
        
        self._minFilterList.append(osg.Texture2D.NEAREST_MIPMAP_NEAREST)
        self._magFilterList.append(osg.Texture2D.LINEAR)
        self._textList.append("nearest mip mapping (default filtering)\nsetFilter(MIN_FILTER,)\nsetFilter(MAG_FILTER,LINEAR)")

        self._minFilterList.append(osg.Texture2D.LINEAR_MIPMAP_NEAREST)
        self._magFilterList.append(osg.Texture2D.LINEAR)
        self._textList.append("bi-linear mip mapping\nsetFilter(MIN_FILTER,LINEAR_MIPMAP_NEAREST)\nsetFilter(MAG_FILTER,LINEAR)")

        self._minFilterList.append(osg.Texture2D.NEAREST_MIPMAP_LINEAR)
        self._magFilterList.append(osg.Texture2D.LINEAR)
        self._textList.append("bi-linear mip mapping\nsetFilter(MIN_FILTER,NEAREST_MIPMAP_LINEAR)\nsetFilter(MAG_FILTER,LINEAR)")

        self.setValues()
    

    def apply(self, node, nv):
        if nv.getFrameStamp():
        
            currTime = nv.getFrameStamp().getReferenceTime()
            if currTime-self._prevTime>self._delay:
            
                # update filter modes and text.
                self.setValues()
                
                # advance the current positon, wrap round if required.
                self._currPos += 1
                if self._currPos>=len(self._minFilterList):
                    self._currPos=0
                
                # record time
                self._prevTime = currTime
    
    def setValues(self):
    
        self._texture.setFilter(osg.Texture2D.MIN_FILTER,self._minFilterList[self._currPos])
        self._texture.setFilter(osg.Texture2D.MAG_FILTER,self._magFilterList[self._currPos])

        self._text.setText(self._textList[self._currPos])
    



def createFilterWall(bb, filename):

    group = osg.Group()
    
    # left hand side of bounding box.
    top_left = osg.Vec3(bb.xMin(),bb.yMin(),bb.zMax())
    bottom_left = osg.Vec3(bb.xMin(),bb.yMin(),bb.zMin())
    bottom_right = osg.Vec3(bb.xMin(),bb.yMax(),bb.zMin())
    top_right = osg.Vec3(bb.xMin(),bb.yMax(),bb.zMax())
    center = osg.Vec3(bb.xMin(),(bb.yMin()+bb.yMax())*0.5,(bb.zMin()+bb.zMax())*0.5)
    height = bb.zMax()-bb.zMin()
    
    # create the geometry for the wall.
    geom = osg.Geometry()
    
    vertices = osg.Vec3Array(4)
    vertices[0] = top_left
    vertices[1] = bottom_left
    vertices[2] = bottom_right
    vertices[3] = top_right
    geom.setVertexArray(vertices)
    
    texcoords = osg.Vec2Array(4)
    texcoords[0].set(0.0,1.0)
    texcoords[1].set(0.0,0.0)
    texcoords[2].set(1.0,0.0)
    texcoords[3].set(1.0,1.0)
    geom.setTexCoordArray(0,texcoords)

    normals = osg.Vec3Array(1)
    normals[0].set(1.0,0.0,0.0)
    geom.setNormalArray(normals)
    geom.setNormalBinding(osg.Geometry.BIND_OVERALL)
    
    colors = osg.Vec4Array(1)
    colors[0].set(1.0,1.0,1.0,1.0)
    geom.setColorArray(colors)
    geom.setColorBinding(osg.Geometry.BIND_OVERALL)

    geom.addPrimitiveSet(osg.DrawArrays(GL_QUADS,0,4))
    
    geom_geode = osg.Geode()
    geom_geode.addDrawable(geom)
    group.addChild(geom_geode)
    
    
    # set up the texture state.    
    texture = osg.Texture2D()
    texture.setDataVariance(osg.Object.DYNAMIC); # protect from being optimized away as static state.
    texture.setImage(osgDB.readImageFile(filename))
    
    stateset = geom.getOrCreateStateSet()
    stateset.setTextureAttributeAndModes(0,texture,osg.StateAttribute.ON)
    
    # create the text label.
    
    text = osgText.Text()
    text.setFont("fonts/arial.ttf")
    text.setPosition(center)
    text.setCharacterSize(height*0.03)
    text.setAlignment(osgText.Text.CENTER_CENTER)
    text.setAxisAlignment(osgText.Text.YZ_PLANE)
    
    text_geode = osg.Geode()
    text_geode.addDrawable(text)
    
    text_stateset = text_geode.getOrCreateStateSet()
    text_stateset.setAttributeAndModes(osg.PolygonOffset(-1.0,-1.0),osg.StateAttribute.ON)
    
    group.addChild(text_geode)

    # set the update callback to cycle through the various min and mag filter modes.
    group.setUpdateCallback(FilterCallback(texture,text))
    return group


#####################################################################/
#
# anisotropic wall and animation callback.
#

class AnisotropicCallback( osg.NodeCallback):
    def __init__(self, texture,text, delay=1.0):
        self._texture = texture
        self._text = text
        self._delay = delay
        self._currPos = 0
        self._prevTime = 0.0
        self._maxAnisotropyList = []
        self._textList = []
    
        self._maxAnisotropyList.append(1.0)
        self._textList.append("No anisotropic filtering (default)\nsetMaxAnisotropy(1.0)")
        
        self._maxAnisotropyList.append(2.0)
        self._textList.append("Anisotropic filtering\nsetMaxAnisotropy(2.0)")
        
        self._maxAnisotropyList.append(4.0)
        self._textList.append("Anisotropic filtering\nsetMaxAnisotropy(4.0)")
        
        self._maxAnisotropyList.append(8.0)
        self._textList.append("Anisotropic filtering\nsetMaxAnisotropy(8.0)")
        
        self._maxAnisotropyList.append(16.0)
        self._textList.append("Higest quality anisotropic filtering\nsetMaxAnisotropy(16.0)")

        self.setValues()
    

    def apply(self, node, nv):
    
        if nv.getFrameStamp():
        
            currTime = nv.getFrameStamp().getReferenceTime()
            if currTime-self._prevTime>self._delay:
            
                # update filter modes and text.
                self.setValues()
                
                # advance the current positon, wrap round if required.
                self._currPos += 1
                if self._currPos>=len(self._maxAnisotropyList): self._currPos=0

                # record time
                self._prevTime = currTime
    
    
    def setValues(self):
        self._texture.setMaxAnisotropy(self._maxAnisotropyList[self._currPos])
        self._text.setText(self._textList[self._currPos])

def createAnisotripicWall(bb, filename):

    group = osg.Group()
    
    # left hand side of bounding box.
    top_left = osg.Vec3(bb.xMin(),bb.yMax(),bb.zMin())
    bottom_left = osg.Vec3(bb.xMin(),bb.yMin(),bb.zMin())
    bottom_right = osg.Vec3(bb.xMax(),bb.yMin(),bb.zMin())
    top_right = osg.Vec3(bb.xMax(),bb.yMax(),bb.zMin())
    center = osg.Vec3((bb.xMin()+bb.xMax())*0.5,(bb.yMin()+bb.yMax())*0.5,bb.zMin());    
    height = bb.yMax()-bb.yMin()
    
    # create the geometry for the wall.
    geom = osg.Geometry()
    
    vertices = osg.Vec3Array(4)
    vertices[0] = top_left
    vertices[1] = bottom_left
    vertices[2] = bottom_right
    vertices[3] = top_right
    geom.setVertexArray(vertices)
    
    texcoords = osg.Vec2Array(4)
    texcoords[0].set(0.0,1.0)
    texcoords[1].set(0.0,0.0)
    texcoords[2].set(1.0,0.0)
    texcoords[3].set(1.0,1.0)
    geom.setTexCoordArray(0,texcoords)

    normals = osg.Vec3Array(1)
    normals[0].set(0.0,0.0,1.0)
    geom.setNormalArray(normals)
    geom.setNormalBinding(osg.Geometry.BIND_OVERALL)
    
    colors = osg.Vec4Array(1)
    colors[0].set(1.0,1.0,1.0,1.0)
    geom.setColorArray(colors)
    geom.setColorBinding(osg.Geometry.BIND_OVERALL)

    geom.addPrimitiveSet(osg.DrawArrays(GL_QUADS,0,4))
    
    geom_geode = osg.Geode()
    geom_geode.addDrawable(geom)
    group.addChild(geom_geode)
    
    
    # set up the texture state.    
    texture = osg.Texture2D()
    texture.setDataVariance(osg.Object.DYNAMIC); # protect from being optimized away as static state.
    texture.setImage(osgDB.readImageFile(filename))
    
    stateset = geom.getOrCreateStateSet()
    stateset.setTextureAttributeAndModes(0,texture,osg.StateAttribute.ON)
    
    # create the text label.
    
    text = osgText.Text()
    text.setFont("fonts/arial.ttf")
    text.setPosition(center)
    text.setCharacterSize(height*0.03)
    text.setColor(osg.Vec4(1.0,0.0,1.0,1.0))
    text.setAlignment(osgText.Text.CENTER_CENTER)
    text.setAxisAlignment(osgText.Text.XY_PLANE)
    
    text_geode = osg.Geode()
    text_geode.addDrawable(text)
    
    text_stateset = text_geode.getOrCreateStateSet()
    text_stateset.setAttributeAndModes(osg.PolygonOffset(-1.0,-1.0),osg.StateAttribute.ON)
    
    group.addChild(text_geode)

    # set the update callback to cycle through the various min and mag filter modes.
    group.setUpdateCallback(AnisotropicCallback(texture,text))

    return group


#####################################################################/
#
# wrap wall and animation callback.
#
class WrapCallback( osg.NodeCallback):
    def __init__(self, texture, text, delay=1.0):
        self._texture = texture
        self._text = text
        self._delay = delay
        self._currPos = 0
        self._prevTime = 0.0
        self._wrapList = []
        self._textList = []

        self._wrapList.append(osg.Texture2D.CLAMP)
        self._textList.append("Default tex coord clamp\nsetWrap(WRAP_S,CLAMP)")
        
        self._wrapList.append(osg.Texture2D.CLAMP_TO_EDGE)
        self._textList.append("Clamp to edge extension\nsetWrap(WRAP_S,CLAMP_TO_EDGE)")
        
        self._wrapList.append(osg.Texture2D.CLAMP_TO_BORDER)
        self._textList.append("Clamp to border color extension\nsetWrap(WRAP_S,CLAMP_TO_BORDER)")
        
        self._wrapList.append(osg.Texture2D.REPEAT)
        self._textList.append("Repeat wrap\nsetWrap(WRAP_S,REPEAT)")
        
        self._wrapList.append(osg.Texture2D.MIRROR)
        self._textList.append("Mirror wrap extension\nsetWrap(WRAP_S,MIRROR)")

        self.setValues()
    

    def apply(self, node, nv):
    
        if nv.getFrameStamp():
        
            currTime = nv.getFrameStamp().getReferenceTime()
            if currTime-self._prevTime>self._delay:
            
                # update filter modes and text.
                self.setValues()
                
                # advance the current positon, wrap round if required.
                self._currPos += 1
                if self._currPos>=len(self._wrapList): self._currPos=0;                    
                
                # record time
                self._prevTime = currTime
    
    def setValues(self):
    
        self._texture.setWrap(osg.Texture2D.WRAP_S,self._wrapList[self._currPos])
        self._texture.setWrap(osg.Texture2D.WRAP_T,self._wrapList[self._currPos])

        self._text.setText(self._textList[self._currPos])
    


def createWrapWall(bb, filename):

    group = osg.Group()
    
    # left hand side of bounding box.
    top_left = osg.Vec3(bb.xMax(),bb.yMax(),bb.zMax())
    bottom_left = osg.Vec3(bb.xMax(),bb.yMax(),bb.zMin())
    bottom_right = osg.Vec3(bb.xMax(),bb.yMin(),bb.zMin())
    top_right = osg.Vec3(bb.xMax(),bb.yMin(),bb.zMax())
    center = osg.Vec3(bb.xMax(),(bb.yMin()+bb.yMax())*0.5,(bb.zMin()+bb.zMax())*0.5)
    height = bb.zMax()-bb.zMin()
    
    # create the geometry for the wall.
    geom = osg.Geometry()
    
    vertices = osg.Vec3Array(4)
    vertices[0] = top_left
    vertices[1] = bottom_left
    vertices[2] = bottom_right
    vertices[3] = top_right
    geom.setVertexArray(vertices)
    
    texcoords = osg.Vec2Array(4)
    texcoords[0].set(-1.0,2.0)
    texcoords[1].set(-1.0,-1.0)
    texcoords[2].set(2.0,-1.0)
    texcoords[3].set(2.0,2.0)
    geom.setTexCoordArray(0,texcoords)

    normals = osg.Vec3Array(1)
    normals[0].set(-1.0,0.0,0.0)
    geom.setNormalArray(normals)
    geom.setNormalBinding(osg.Geometry.BIND_OVERALL)
    
    colors = osg.Vec4Array(1)
    colors[0].set(1.0,1.0,1.0,1.0)
    geom.setColorArray(colors)
    geom.setColorBinding(osg.Geometry.BIND_OVERALL)

    geom.addPrimitiveSet(osg.DrawArrays(GL_QUADS,0,4))
    
    geom_geode = osg.Geode()
    geom_geode.addDrawable(geom)
    group.addChild(geom_geode)
    
    
    # set up the texture state.    
    texture = osg.Texture2D()
    texture.setDataVariance(osg.Object.DYNAMIC); # protect from being optimized away as static state.
    texture.setBorderColor(osg.Vec4(1.0,1.0,1.0,0.5)); # only used when wrap is set to CLAMP_TO_BORDER
    texture.setImage(osgDB.readImageFile(filename))
    
    stateset = geom.getOrCreateStateSet()
    stateset.setTextureAttributeAndModes(0,texture,osg.StateAttribute.ON)
    stateset.setMode(GL_BLEND,osg.StateAttribute.ON)
    stateset.setRenderingHint(osg.StateSet.TRANSPARENT_BIN)
    
    # create the text label.
    
    text = osgText.Text()
    text.setFont("fonts/arial.ttf")
    text.setPosition(center)
    text.setCharacterSize(height*0.03)
    text.setAlignment(osgText.Text.CENTER_CENTER)
    text.setAxisAlignment(osgText.Text.YZ_PLANE)
    
    text_geode = osg.Geode()
    text_geode.addDrawable(text)
    
    text_stateset = text_geode.getOrCreateStateSet()
    text_stateset.setAttributeAndModes(osg.PolygonOffset(-1.0,-1.0),osg.StateAttribute.ON)
    
    group.addChild(text_geode)

    # set the update callback to cycle through the various min and mag filter modes.
    group.setUpdateCallback(WrapCallback(texture,text))
    
    return group
    


#####################################################################/
#
# sublooad wall and animation callback.
#

class ImageUpdateCallback( osg.NodeCallback):

    def __init__(self, texture,text, delay=1.0):
        self._texture = texture
        self._text = text
        self._delay = delay
        self._currPos = 0
        self._prevTime = 0.0
        self._imageList = []
        self._textList = []
    
        self._imageList.append(osgDB.readImageFile("Images/dog_left_eye.jpg"))
        self._textList.append("Subloaded Image 1 - dog_left_eye.jpg")
        
        self._imageList.append(osgDB.readImageFile("Images/dog_right_eye.jpg"))
        self._textList.append("Subloaded Image 2 - dog_right_eye.jpg")
        
        self.setValues()
    

    def apply(self, node, nv):
        if nv.getFrameStamp():
            currTime = nv.getFrameStamp().getReferenceTime()
            if currTime-self._prevTime>self._delay:
            
                # update filter modes and text.
                self.setValues()
                
                # advance the current positon, wrap round if required.
                self._currPos += 1
                if self._currPos>=len(self._imageList): self._currPos=0;                    
                
                # record time
                self._prevTime = currTime
    
    def setValues(self):
    
        # Note, as long as the images are the same dimensions subloading will be used
        # to update the textures.  If dimensions change then the texture objects have
        # to be deleted and re-recreated.
        #
        # The load/subload happens during the draw traversal so doesn't happen on
        # the setImage which just updates internal pointers and modifed flags.

        self._texture.setImage(self._imageList[self._currPos])
        
        #_texture.dirtyTextureObject()

        self._text.setText(self._textList[self._currPos])

def createSubloadWall(bb):

    group = osg.Group()
    
    # left hand side of bounding box.
    top_left = osg.Vec3(bb.xMin(),bb.yMax(),bb.zMax())
    bottom_left = osg.Vec3(bb.xMin(),bb.yMax(),bb.zMin())
    bottom_right = osg.Vec3(bb.xMax(),bb.yMax(),bb.zMin())
    top_right = osg.Vec3(bb.xMax(),bb.yMax(),bb.zMax())
    center = osg.Vec3((bb.xMax()+bb.xMin())*0.5,bb.yMax(),(bb.zMin()+bb.zMax())*0.5)
    height = bb.zMax()-bb.zMin()
    
    # create the geometry for the wall.
    geom = osg.Geometry()
    
    vertices = osg.Vec3Array(4)
    vertices[0] = top_left
    vertices[1] = bottom_left
    vertices[2] = bottom_right
    vertices[3] = top_right
    geom.setVertexArray(vertices)
    
    texcoords = osg.Vec2Array(4)
    texcoords[0].set(0.0,1.0)
    texcoords[1].set(0.0,0.0)
    texcoords[2].set(1.0,0.0)
    texcoords[3].set(1.0,1.0)
    geom.setTexCoordArray(0,texcoords)

    normals = osg.Vec3Array(1)
    normals[0].set(0.0,-1.0,0.0)
    geom.setNormalArray(normals)
    geom.setNormalBinding(osg.Geometry.BIND_OVERALL)
    
    colors = osg.Vec4Array(1)
    colors[0].set(1.0,1.0,1.0,1.0)
    geom.setColorArray(colors)
    geom.setColorBinding(osg.Geometry.BIND_OVERALL)

    geom.addPrimitiveSet(osg.DrawArrays(GL_QUADS,0,4))
    
    geom_geode = osg.Geode()
    geom_geode.addDrawable(geom)
    group.addChild(geom_geode)
    
    
    # set up the texture state.    
    texture = osg.Texture2D()
    texture.setDataVariance(osg.Object.DYNAMIC); # protect from being optimized away as static state.
    texture.setFilter(osg.Texture2D.MIN_FILTER,osg.Texture2D.LINEAR)
    texture.setFilter(osg.Texture2D.MAG_FILTER,osg.Texture2D.LINEAR)
    
    stateset = geom.getOrCreateStateSet()
    stateset.setTextureAttributeAndModes(0,texture,osg.StateAttribute.ON)
    
    # create the text label.
    
    text = osgText.Text()
    text.setFont("fonts/arial.ttf")
    text.setPosition(center)
    text.setCharacterSize(height*0.03)
    text.setAlignment(osgText.Text.CENTER_CENTER)
    text.setAxisAlignment(osgText.Text.XZ_PLANE)
    
    text_geode = osg.Geode()
    text_geode.addDrawable(text)
    
    text_stateset = text_geode.getOrCreateStateSet()
    text_stateset.setAttributeAndModes(osg.PolygonOffset(-1.0,-1.0),osg.StateAttribute.ON)
    
    group.addChild(text_geode)

    # set the update callback to cycle through the various min and mag filter modes.
    group.setUpdateCallback(ImageUpdateCallback(texture,text))
    
    return group
    



def createModel():


    # create the root node which will hold the model.
    root = osg.Group()
    
    # turn off lighting 
    root.getOrCreateStateSet().setMode(GL_LIGHTING,osg.StateAttribute.OFF)

    bb = osg.BoundingBox(0.0,0.0,0.0,1.0,1.0,1.0)

    root.addChild(createFilterWall(bb,"Images/lz.rgb"))
    root.addChild(createAnisotripicWall(bb,"Images/primitives.gif"))
    root.addChild(createWrapWall(bb,"Images/tree0.rgba"))
    root.addChild(createSubloadWall(bb))
    
    return root


def main(argv):

    # use an ArgumentParser object to manage the program arguments.
    arguments = osg.ArgumentParser(argv)

    # set up the usage document, in case we need to print out how to use this program.
    arguments.getApplicationUsage().setDescription(arguments.getApplicationName()+" is the example which demonstrates use of 2D textures.")
    arguments.getApplicationUsage().setCommandLineUsage(arguments.getApplicationName()+" [options] filename ...")
    arguments.getApplicationUsage().addCommandLineOption("-h or --help","Display this information")
   
    # construct the viewer.
    viewer = osgProducer.Viewer(arguments)

    # set up the value with sensible default event handlers.
    viewer.setUpViewer(osgProducer.Viewer.STANDARD_SETTINGS)

    # get details on keyboard and mouse bindings used by the viewer.
    viewer.getUsage(arguments.getApplicationUsage())

    # if user request help write it out to cout.
    if arguments.read("-h") or arguments.read("--help"):
    
        arguments.getApplicationUsage().write()
        return 1
    

    # any option left unread are converted into errors to write out later.
    arguments.reportRemainingOptionsAsUnrecognized()

    # report any errors if they have occured when parsing the program aguments.
    if arguments.errors():
    
        arguments.writeErrorMessages()
        return 1
    

    # create a model from the images.
    rootNode = createModel()

    # add model to viewer.
    viewer.setSceneData( rootNode )

    # create the windows and run the threads.
    viewer.realize()

    while not viewer.done():
    
        # wait for all cull and draw threads to complete.
        viewer.sync()

        # update the scene by traversing it with the the update visitor which will
        # call all node update callbacks and animations.
        viewer.update()

        # fire off the cull and draw traversals of the scene.
        viewer.frame()

    

    # wait for all cull and draw threads to complete before exit.
    viewer.sync()
    return 0

if __name__ == "__main__":
    main(sys.argv)
