// Copyright (C) 2002-2003 Gideon May (gideon@computer.org)
//
// Permission to copy, use, sell and distribute this software is granted
// provided this copyright notice appears in all copies.
// Permission to modify the code and to distribute modified code is granted
// provided this copyright notice appears in all copies, and a notice
// that the code was modified is included with the copyright notice.
//
// This software is provided "as is" without express or implied warranty,
// and with no claim as to its suitability for any purpose.

#include <boost/python/class.hpp>
#include <boost/python/scope.hpp>
#include <boost/python/call_method.hpp>
#include <boost/python/object.hpp>
#include <boost/python/extract.hpp>
#include <boost/python/return_value_policy.hpp>

#include <osg/Texture1D>
#include <osg/State>

#include "StateAttribute.hpp"
#include "held_ptr.hpp"

using namespace boost::python;

namespace {
osg::Texture1D * asTexture1D(osg::StateAttribute * self) { return dynamic_cast<osg::Texture1D*>(self); }

class SubloadCallback : public osg::Texture1D::SubloadCallback {
  public:
    SubloadCallback(PyObject * p) : _self(p) {}
    ~SubloadCallback() {}

    virtual void load(const osg::Texture1D& texture,osg::State& state) const{
        try {
            call_method<void>(_self, "load", ptr(&texture), ptr(&state));
        } catch(...) {
            handle_exception();
            PyErr_Print();
            throw_error_already_set();
            return;
        } 
    }
    virtual void subload(const osg::Texture1D& texture,osg::State& state) const{
        try {
            call_method<void>(_self, "subload", ptr(&texture), ptr(&state));
        } catch(...) {
            handle_exception();
            PyErr_Print();
            throw_error_already_set();
            return;
        } 
    }
  private:
      PyObject * _self;
};

void setSubloadCallback(osg::Texture1D * self, PyObject * cb)
{
    bool is_callback = extract<osg::Texture1D::SubloadCallback *>(cb).check();

    Py_XINCREF(cb);
    if (is_callback) {
        extract<osg::Texture1D::SubloadCallback *> callback(cb);
        self->setSubloadCallback(callback);
    } else {
        SubloadCallback * callback = new SubloadCallback(cb);
        self->setSubloadCallback(callback);
    }
}
}


namespace PyOSG {
void init_Texture1D()
{
    register_castStateAttribute("asTexture1D", &asTexture1D);

    class_<osg::Texture1D, osg::ref_ptr<osg::Texture1D>, bases<osg::Texture> > texture1d("Texture1D", no_init);

    scope texture1d_scope(texture1d);

    texture1d
         .def(init<>())
         .def("setImage", (void (osg::Texture1D::*)(osg::Image *)) &osg::Texture1D::setImage)
         .def("setImage", (void (osg::Texture1D::*)(unsigned int, osg::Image *)) &osg::Texture1D::setImage)
         .def("getImage",
            (osg::Image *(osg::Texture1D::*)())&osg::Texture1D::getImage,
            return_value_policy<manage_osg_object>())
         .def("setTextureWidth", &osg::Texture1D::setTextureWidth)
         .def("getTextureWidth", &osg::Texture1D::getTextureWidth)
         .def("setNumMipmapLevels", &osg::Texture1D::setNumMipmapLevels)
         .def("getNumMipmapLevels", &osg::Texture1D::getNumMipmapLevels)
         .def("copyTexImage1D", &osg::Texture1D::copyTexImage1D)
         .def("copyTexSubImage1D", &osg::Texture1D::copyTexSubImage1D)
         .def("setSubloadCallback", &setSubloadCallback)
         ;

    class_<osg::Texture1D::SubloadCallback, osg::ref_ptr<SubloadCallback>, bases<osg::Referenced>, boost::noncopyable>("SubloadCallback", no_init)
        .def(init<>())
        ;
}
} // namespace PyOSG
