// Copyright (C) 2002-2003 Gideon May (gideon@computer.org)
//
// Permission to copy, use, sell and distribute this software is granted
// provided this copyright notice appears in all copies.
// Permission to modify the code and to distribute modified code is granted
// provided this copyright notice appears in all copies, and a notice
// that the code was modified is included with the copyright notice.
//
// This software is provided "as is" without express or implied warranty,
// and with no claim as to its suitability for any purpose.

#include <boost/python/class.hpp>
#include <boost/python/scope.hpp>
#include <boost/python/call_method.hpp>
#include <boost/python/object.hpp>
#include <boost/python/extract.hpp>
#include <boost/python/return_value_policy.hpp>

#include <osg/Texture2D>
#include <osg/State>

#include "StateAttribute.hpp"
#include "held_ptr.hpp"

using namespace boost::python;

namespace {
osg::Texture2D * asTexture2D(osg::StateAttribute * self) { return dynamic_cast<osg::Texture2D*>(self); }

class SubloadCallback : public osg::Texture2D::SubloadCallback {
  public:
    SubloadCallback(PyObject * p) : _self(p) {}
    ~SubloadCallback() {}

    virtual void load(const osg::Texture2D& texture,osg::State& state) const{
        try {
            call_method<void>(_self, "load", ptr(&texture), ptr(&state));
        } catch(...) {
            handle_exception();
            PyErr_Print();
            throw_error_already_set();
            return;
        } 
    }
    virtual void subload(const osg::Texture2D& texture,osg::State& state) const{
        try {
            call_method<void>(_self, "subload", ptr(&texture), ptr(&state));
        } catch(...) {
            handle_exception();
            PyErr_Print();
            throw_error_already_set();
            return;
        } 
    }
  private:
      PyObject * _self;
};

void setSubloadCallback(osg::Texture2D * self, PyObject * cb)
{
    bool is_callback = extract<osg::Texture2D::SubloadCallback *>(cb).check();

    Py_XINCREF(cb);
    if (is_callback) {
        extract<osg::Texture2D::SubloadCallback *> callback(cb);
        self->setSubloadCallback(callback);
    } else {
        SubloadCallback * callback = new SubloadCallback(cb);
        self->setSubloadCallback(callback);
    }
}
}

namespace PyOSG {
void init_Texture2D()
{
    register_castStateAttribute("asTexture2D", &asTexture2D);

    class_<osg::Texture2D, osg::ref_ptr<osg::Texture2D>, bases<osg::Texture> > texture2d("Texture2D", no_init);

    scope texture2d_scope(texture2d);

    texture2d
        .def(init<>())
        .def("setImage", 
            (void (osg::Texture2D::*)(osg::Image *))
            &osg::Texture2D::setImage,
            "Set the texture image."
            )
        .def("getImage", 
            (osg::Image *(osg::Texture2D::*)())
            &osg::Texture2D::getImage, 
            return_value_policy<manage_osg_object>(),
            "Get the texture image."
            )
        .def("setTextureSize", 
            &osg::Texture2D::setTextureSize,
            "Set the texture width and height. If width or height are zero then"
            "the repsective size value is calculated from the source image sizes."
            )

        .def("getTextureWidth", 
            &osg::Texture2D::getTextureWidth)
        .def("getTextureHeight", 
            &osg::Texture2D::getTextureHeight)

        .def("setSubloadCallback", 
            &setSubloadCallback)

        .def("setNumMipmapLevels", 
            &osg::Texture2D::setNumMipmapLevels)
        .def("getNumMipmapLevels", 
            &osg::Texture2D::getNumMipmapLevels)

        .def("copyTexImage2D", 
            &osg::Texture2D::copyTexImage2D)
        .def("copyTexSubImage2D", 
            &osg::Texture2D::copyTexSubImage2D)
         ;

    class_<osg::Texture2D::SubloadCallback, osg::ref_ptr<SubloadCallback>, bases<osg::Referenced>, boost::noncopyable>("SubloadCallback", no_init)
        .def(init<>())
        ;
}
} // namespace PyOSG
