// Copyright (C) 2002-2003 Gideon May (gideon@computer.org)
//
// Permission to copy, use, sell and distribute this software is granted
// provided this copyright notice appears in all copies.
// Permission to modify the code and to distribute modified code is granted
// provided this copyright notice appears in all copies, and a notice
// that the code was modified is included with the copyright notice.
//
// This software is provided "as is" without express or implied warranty,
// and with no claim as to its suitability for any purpose.

#include "osgScript/PythonScript"

#include <osgDB/Registry>
#include <osgDB/Input>
#include <osgDB/Output>

#include <osg/Notify>
#include <iostream>
#include <sstream>

using namespace osg;
using namespace osgDB;
using namespace osgScript;

// forward declare functions to use later.
bool PythonScript_readLocalData(Object& obj, Input& fr);
bool PythonScript_writeLocalData(const Object& obj, Output& fw);

osgDB::RegisterDotOsgWrapperProxy PythonScript_Proxy
(
    new PythonScript,
    "PythonScript",
    "Object Node Group Script PythonScript",
    &PythonScript_readLocalData,
    &PythonScript_writeLocalData
);

bool PythonScript_readLocalData(Object& obj, Input& fr)
{
    bool iteratorAdvanced = false;

    PythonScript& script = static_cast<PythonScript&>(obj);

    if (fr.matchSequence("updateFunc %s")) {
        script.setUpdateFunc(fr[1].getStr());
        fr += 2;
        iteratorAdvanced = true;
    } else if (fr.matchSequence("postUpdateFunc %s")) {
        script.setPostUpdateFunc(fr[1].getStr());
        fr += 2;
        iteratorAdvanced = true;
    } else if (fr.matchSequence("cullFunc %s")) {
        script.setCullFunc(fr[1].getStr());
        fr += 2;
        iteratorAdvanced = true;
    } else if (fr.matchSequence("postCullFunc %s")) {
        script.setPostCullFunc(fr[1].getStr());
        fr += 2;
        iteratorAdvanced = true;
    } else if (fr.matchSequence("initFunc %s")) {
        script.setInitFunc(fr[1].getStr());
        fr += 2;
        iteratorAdvanced = true;
    } else if (fr.matchSequence("code {")) {
        std::string code;
        fr += 2;
        iteratorAdvanced = true;
        int entry = fr[0].getNoNestedBrackets();
        while (!fr.eof() && fr[0].getNoNestedBrackets() >= entry) {
            if (fr[0].getStr()) {
            code.append(std::string(fr[0].getStr()));
            code += '\n' ;
            }
            ++fr;
        }
        script.setInlineCode(code.c_str());
        fr += 2;
        iteratorAdvanced = true;
    } else if (fr.matchSequence("script %s")) {
        script.setExternCode(fr[1].getStr());
        fr += 2;
        iteratorAdvanced = true;
    }
    return iteratorAdvanced;
}


bool PythonScript_writeLocalData(const Object& obj, Output& fw)
{
    const PythonScript& script = static_cast<const PythonScript&>(obj);

    std::cerr << "PythonScript_writeLocalData\n";

    if (!script.getUpdateFunc().empty()) {
        fw.indent() << "updateFunc \"" << script.getUpdateFunc() << "\"" << std::endl;
    }

    if (!script.getPostUpdateFunc().empty()) {
        fw.indent() << "postUpdateFunc \"" << script.getPostUpdateFunc() << "\"" << std::endl;
    }

    if (!script.getCullFunc().empty()) {
        fw.indent() << "cullFunc \"" << script.getCullFunc() << "\"" << std::endl;
    }

    if (!script.getPostCullFunc().empty()) {
        fw.indent() << "postCullFunc \"" << script.getPostCullFunc() << "\"" << std::endl;
    }

    if (!script.getInitFunc().empty()) {
        fw.indent() << "initFunc \"" << script.getInitFunc() << "\"" << std::endl;
    }

    if (!script.getInlineCode().empty()) {
        std::vector<std::string> lines;
        std::istringstream iss(script.getInlineCode());
        std::string line;
        while (std::getline(iss, line)) lines.push_back(line);

        fw.indent() << "code { \n";
        fw.moveIn();

        std::vector<std::string>::const_iterator j;
        for (j=lines.begin(); j!=lines.end(); ++j) {
            fw.indent() << fw.wrapString(*j) << "\n";
        }

        fw.moveOut();
        fw.indent() << "}\n";
    }

    if (!script.getExternCode().empty()) {
        fw.indent() << "script \"" << script.getExternCode() << "\"" << std::endl;
    }

    return true;
}

